/*
 * dnsblock.c
 *
 * Copyright (C) 2004 Andreas Beck <becka@bedatec.de>                           
 *                                                                              
 * This program is free software; you can redistribute it and/or modify it      
 * under the terms of the GNU General Public License as published by the        
 * Free Software Foundation; either version 2 of the License, or (at your       
 * option) any later version.                                                   
 *                                                                              
 * This program is distributed in the hope that it will be useful, but          
 * WITHOUT ANY WARRANTY; without even the implied warranty of                   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General    
 * Public License for more details.                                             
 *                                                                              
 * You should have received a copy of the GNU General Public License along      
 * with this program; if not, write to the Free Software Foundation, Inc.,      
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                        
 *                                                                              
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <pwd.h>
#include <sys/types.h>

#include "config.h"

struct parms {
	int port;
	char *bindaddr;
	char *fakeaddr;
	char *mylog;
	char *runas;
} parms= {
	53,
	NULL,
	NULL,
	MYLOG,
	NULL
};

struct mydata {
	FILE *mylog;
	const char *prgnam;
	char hostname[256];
	int	mypid;
} mydata = {
	NULL,
	"",
	"",
	0
};

static int init_mydata(char *prgnam) {

	int logfd;
	
	mydata.prgnam=prgnam;
	if (strrchr(mydata.prgnam,'/')) mydata.prgnam=strrchr(mydata.prgnam,'/')+1;

	if (-1==(logfd=open(parms.mylog,O_CREAT|O_APPEND|O_WRONLY,S_IRUSR|S_IWUSR))) {
		fprintf(stderr,"Cannot open logfile.Aborting.\n");
		return 1;
	}
	if (NULL==(mydata.mylog=fdopen(logfd,"a"))) {
		close(logfd);
		fprintf(stderr,"Cannot fdopen logfile.Aborting.\n");
		return 1;
	}
	dup2(logfd,1);
	dup2(logfd,2);
	close(0);
	gethostname(mydata.hostname,sizeof(mydata.hostname));
	mydata.mypid=(int)getpid();
	return 0;
}

static int exit_mydata(void) {

	fclose(mydata.mylog);
	return 0;
}

static char *logdate(void) {
	static char timebuf[256];
	struct tm *tm;
	time_t tim;
	time(&tim);
	tm=localtime(&tim);
	strcpy(timebuf,asctime(tm)+4);
	timebuf[15]='\0';
	return timebuf;
}

int dolog(char *format,...) {
	va_list ap;
	int rc;

	va_start(ap,format);

	fprintf(mydata.mylog,"%s %s %s[%d]: ",
		logdate(),mydata.hostname,mydata.prgnam,(int)mydata.mypid);

	rc=vfprintf(mydata.mylog,format,ap);
	fprintf(mydata.mylog,"\n");
	fflush(mydata.mylog);

	va_end(ap);
	return rc;
}

int drop_root(void) {
	struct passwd *pwd;

	if (parms.runas) {
		pwd=getpwnam(parms.runas);
		setresuid(pwd->pw_uid,pwd->pw_uid,pwd->pw_uid);
	}
}

int dowatch(void) {

	int sock,rc,pas,x;
	int udpprotonum;
	struct protoent *udpproto;
	struct hostent  *hentry;
	struct sockaddr_in listenaddr;
	struct sockaddr_in peeraddr;
	struct sockaddr_in fakeaddr;
	unsigned char buffer[1024];
	
	udpproto=getprotobyname("udp");
	udpprotonum=udpproto->p_proto;

	fakeaddr.sin_family=AF_INET;
	fakeaddr.sin_addr.s_addr=htonl(INADDR_ANY);
	fakeaddr.sin_port=htons(parms.port);
	if (parms.fakeaddr) {
		if (NULL==(hentry=gethostbyname(parms.fakeaddr))) {
			dolog("Can't resolve address to fake to.");
			return 1;
		}
		memcpy(&fakeaddr.sin_addr.s_addr,hentry->h_addr,sizeof(fakeaddr.sin_addr.s_addr));
	}
	
	sock=socket(PF_INET,SOCK_DGRAM,udpprotonum);
	if (sock==-1) {
		dolog("Can't allocate a UDP socket.");
		return 1;
	}
	
	listenaddr.sin_family=AF_INET;
	listenaddr.sin_addr.s_addr=htonl(INADDR_ANY);
	listenaddr.sin_port=htons(parms.port);
	if (parms.bindaddr) {
		if (NULL==(hentry=gethostbyname(parms.bindaddr))) {
			dolog("Can't resolve address to bind to.");
			return 1;
		}
		memcpy(&listenaddr.sin_addr.s_addr,hentry->h_addr,sizeof(listenaddr.sin_addr.s_addr));
	}
	if (bind(sock,(struct sockaddr *)&listenaddr,sizeof(listenaddr))) {
		dolog("Can't bind to given IP/Port.");
		return 1;
	}
	if (parms.runas) {
		if (drop_root()) {
			dolog("Can't drop to given uid.");
			return 1;
		}
	}

	dolog("Startup o.k. - waiting for clients to connect.");

	while(1) {
		memset(buffer,0,sizeof(buffer));
		pas=sizeof(peeraddr);
		rc=recvfrom(sock,buffer,sizeof(buffer),0,(struct sockaddr *)&peeraddr,&pas);
		if (rc<=0) {
			dolog("Got illegal packet or interrupt. Ignoring.");
			continue;
		}
		if (rc<12) {
			dolog("Packet too small. Ignoring.");
			continue;
		}
		if ((buffer[2]&0xfe)!=0) {
			dolog("Not a standard query. Ignoring.");
			continue;
		}
		if (buffer[ 4]!=0||buffer[ 5]!=1||
		    buffer[ 6]!=0||buffer[ 7]!=0||
		    buffer[ 8]!=0||buffer[ 9]!=0||
		    buffer[10]!=0||buffer[11]!=0) {
		    	dolog("Expect only simple single queries. Ignoring.");
		    	continue;
		}
		for(x=12;x<rc;x++) {
			if (buffer[x]) x+=buffer[x];
			else {
				x++;
				break;
			}
		}
		if (x+4>rc) {
			dolog("Illegal packet. Attack?");
			continue;
		}
		if (buffer[x]!=0||buffer[x+1]!=1) {
			dolog("Not an A Query. Ignoring.");
			continue;
		}
		if (buffer[x+2]!=0||buffer[x+3]!=1) {
			dolog("Not an INET Class query. Ignoring.");
			continue;
		}
		x+=4;
		if (rc!=x) {
			dolog("Packet length mismatch got %d should be %d.",rc,x);
			continue;
		}
		if (x+12+sizeof(fakeaddr.sin_addr.s_addr)>sizeof(buffer)) {
			dolog("Answer would exceed buffer. Ignoring.");
			continue;
		}
		/* O.k- Query is acceptable. Build an answer. */
		buffer[2]|=0x80;	/* Answer*/
		buffer[3]|=0x80;	/* can do recursive */
		buffer[3]&=~0x0f;	/* No error */

		buffer[6]=0;		/* We have one answer block */
		buffer[7]=1;
		
		buffer[x]  =0xc0;	/* NAME - Reference to */
		buffer[x+1]=12;		/* Position 12, the query */

		buffer[x+2]=0;		/* Type = Host address */
		buffer[x+3]=1;
		
		buffer[x+4]=0;		/* Class = INET */
		buffer[x+5]=1;
		
		/* TTL = 0 - our reply is fake anyway. */
		buffer[x+6]=buffer[x+7]=buffer[x+8]=buffer[x+9]=0;
		
		buffer[x+10]=0;
		buffer[x+11]=sizeof(fakeaddr.sin_addr.s_addr);
		
		memcpy(&buffer[x+12],&fakeaddr.sin_addr.s_addr,sizeof(fakeaddr.sin_addr.s_addr));
		rc=x+12+sizeof(fakeaddr.sin_addr.s_addr);
		
		sendto(sock,&buffer,rc,0,(struct sockaddr *)&peeraddr,pas);
		dolog("Sent answer to %d.%d.%d.%d.",
			( peeraddr.sin_addr.s_addr     &0xff),
			((peeraddr.sin_addr.s_addr>> 8)&0xff),
			((peeraddr.sin_addr.s_addr>>16)&0xff),
			((peeraddr.sin_addr.s_addr>>24)&0xff)
			);
	}
	
	close(sock);
	
	return 0;
}

int daemonize(void) {


	switch(fork()) {
		case -1:
			fprintf(stderr,"Cannot fork to become daemon.");
			return -1;
		case 0: 
			setsid();
			return 0;
		default:
			return 1;
	}
}

int main(int argc, char **argv) {

	int optchar;
	int daemon=0;

	while(-1!=(optchar=getopt(argc,argv,"Dl:p:L:f:r:"))) {
		switch(optchar) {
			case 'D': daemon=1;break;
			case 'l': parms.bindaddr=optarg;break;
			case 'p': parms.port    =atoi(optarg);break;
			case 'L': parms.mylog   =optarg;break;
			case 'f': parms.fakeaddr=optarg;break;
			case 'r': parms.runas   =optarg;break;
			default: fprintf(stderr,"Usage: %s [-D] [-l listenaddr] [-p listenport] [-L logfilename] \n"
						"       [-f fake address to return] [-r run-as-username]\n",argv[0]);
				return 1;
		}
	}

	if (daemon) {
		switch(daemonize()) {
			case 0: break;
			case 1: return 0;
			default: return 1;
		}
	}
	if (init_mydata(argv[0])) return 1;
	dolog("Starting service");

	dowatch();

	dolog("Stopping Service");
	exit_mydata();
	return 0;
}
